#include <errno.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <string.h>

#include "diffTools.h"


static dt_modifInfo_t dt_MODIF_INFO[] = { 	{ 'd', DEL, "deleted" },
											{ 'c', CHG, "changed" },
											{ 'a', ADD, "added" },
											{ 0, 0, 0 }
								};


static char *dt_RCSDIFF_CMD = "rcsdiff -n %s %s %s";

static dt_modif_t *dt_firstModif=0;
static dt_modif_t *dt_currentModif=0;


void
dt_init()
{
	
	if(dt_firstModif)
	{
		dt_modif_t *mptr = dt_firstModif;
		dt_modif_t *msvg;
		while(mptr)
		{
			if( mptr->firstLine )
			{
				dt_linkedLine_t *lptr = mptr->firstLine;
				dt_linkedLine_t *lsvg;
				while(lptr)
				{
					if(lptr->line)
						free( lptr->line );
					lsvg=lptr;
					lptr=lptr->next;
					free( lsvg );
				}
			
			}
		
			msvg=mptr;
			mptr=mptr->next;
			free( msvg );
		}
		dt_firstModif=dt_currentModif=0;
	}


}

void
dt_printModif( dt_modif_t *ptr )
{	
	int i=0;
	dt_linkedLine_t *lptr = ptr->firstLine;
	printf( "Modif: type=%s start=%d end-%d nb=%d\n", dt_MODIF_INFO[ptr->type].name, ptr->start, ptr->end, ptr->nbLine );
	/*while( lptr )
	{
		printf( "Line %d: %s", i, lptr->line );
		lptr = lptr->next;
		i++;
	}*/
}

int
dt_flag()
{
	return dt_firstModif?1:0;
}

dt_modif_t *
dt_isModif( int id )
{
	dt_modif_t *mptr = dt_firstModif;

/* 	printf( "dt_isModif %d ", id ); */

  	while(mptr)
	{
		if( id<mptr->start )
		{
			/*printf( "CC\n");*/
			return 0;
		}
		/*dt_printModif( mptr );*/

			
		if( (id>=mptr->start) && (id<=mptr->end) )
		{
			/*printf( "OK %d\n", mptr->type );*/
			return mptr;
		}
		mptr=mptr->next;
	}
	/*printf( "NOK\n" );*/
	return NULL;
}


void
dt_addLinkedLine(  dt_modif_t *modif, char *line )
{
	dt_linkedLine_t *ptr = (dt_linkedLine_t *)malloc( sizeof(dt_linkedLine_t) );
	memset( ptr, 0, sizeof(dt_linkedLine_t) );
	
	ptr->line = strdup( line );
	
	if( modif->currentLine )
	{
		ptr->prev = modif->currentLine;
		modif->currentLine->next = ptr;	
		modif->currentLine = ptr;
	}
	else
	{
		modif->currentLine = modif->firstLine = ptr;
	}

}


dt_modif_t *
dt_newModif( dt_modifType_t type, int st, int len )
{
	dt_modif_t *ptr = (dt_modif_t *)malloc( sizeof(dt_modif_t) );
	memset( ptr, 0, sizeof(dt_modif_t) );
	
	ptr->type = type;
	ptr->start = st;
	ptr->nbLine = len;
	ptr->end = ptr->start+len-1;
	
	if( dt_currentModif )
	{
		ptr->prev = dt_currentModif;
		dt_currentModif->next = ptr;
		dt_currentModif = ptr;	
	}
	else
	{
		dt_currentModif=dt_firstModif=ptr;
	}
	
	return ptr;
}

int
dt_checkDiff( char *fname, char *V1, char *V2 )
{
	FILE *pop;
	char lineBuf[4096];
	char cmd[100];
	char strV1[100];
	char strV2[100];
	int pos;
	int nb;
	int i;
	int delCnt=0;
	int addCnt=0;
	
	dt_init();

	sprintf( strV1, "-r%s", V1 );
	sprintf( strV2, "-r%s", V2?V2:"" );


	sprintf( cmd, dt_RCSDIFF_CMD, strV1, V2?strV2:"", fname );

	if( (pop = popen( cmd, "r" )) != NULL )
	{
		while( fgets( lineBuf, 4096, pop ) != NULL )
		{
			switch( lineBuf[0] )
			{
				case 'd':
				{
					sscanf( lineBuf+1, "%d %d", &pos, &nb );
					dt_newModif( DEL, pos-delCnt+addCnt, nb );
					delCnt+=nb;
				}
				break;
				case 'a':
				{
					dt_modif_t *addModif;
					sscanf( lineBuf+1, "%d %d", &pos, &nb );
					addModif=dt_newModif( ADD, pos-delCnt+addCnt+1, nb );
					for( i=0; i<nb; i++ )
					{
						if( fgets( lineBuf, 4096, pop ) != NULL )
						{
							dt_addLinkedLine(  addModif, lineBuf );
						}
					}
					addCnt+=nb;
				}
				break;
				default:
					printf( "Error, No ctrl character found %c !!!", lineBuf[0] );
				break;
					
			
			}
		}
		
		
		pclose(pop);
	}
}

dt_main( int argc, char *argv[] )
{
	dt_modif_t *ptr=0;


	dt_checkDiff( argv[1], argv[2], argv[3] );

	ptr=dt_firstModif;
	while( ptr )
	{
		dt_printModif( ptr );
		ptr = ptr->next;
	}

}
